"""
The Standard Workflow

The standard workflow reads a PPR and executes the following steps:

1. Fetch data
2. Run CASA
3. Deliver results
4. Ingest

The standard workflow is actually structured as a dependency graph. Raw
data is a dependency for running CASA, and the dependency is satisfied by
running the data fetcher. Everything runs in a recovery mode, as if the task
crashed some earlier time and is being rerun.
"""

class Node:
    """
    A node in the dependency graph.
    """

    def __init__(self, action=lambda: True):
        self.action = action
        self.dependencies = []

    def depends_on(self, other_node):
        """Add a dependency to this action."""
        self.dependencies.append(other_node)

    def execute(self):
        """
        Executes this node, if the dependencies all return true.
        Returns True if the execution is successful.
        """
        for dependency in self.dependencies:
            if not dependency.execute():
                return False

        # if we made it here, execute this action
        return self.action()


class PPR:
    @classmethod
    def parse(cls, ppr_filename):
        "Parse the PPR XML file"
        pass

    def build_fetch_step(self): pass
    def build_casa_step(self): pass
    def build_delivery_step(self): pass
    def build_ingest_step(self): pass

class Plan:
    def __init__(self):
        self.ppr = PPR.parse("PPR.xml")

    def generate_plan(self):
        "Read the XML for the PPR and generate the plan from it"
        fetch_step      = Node(self.ppr.build_fetch_step())
        processing_step = Node(self.ppr.build_casa_step())
        delivery_step   = Node(self.ppr.build_delivery_step())
        ingest_step     = Node(self.ppr.build_ingest_step())

        processing_step.depends_on(fetch_step)
        delivery_step  .depends_on(processing_step)
        ingest_step    .depends_on(processing_step)

        goal = Node()
        goal.depends_on(delivery_step)
        goal.depends_on(ingest_step)

        return goal

if __name__ == '__main__':
    plan = Plan()
    plan.generate_plan().execute()
