"""
Finding missing data in the archive.

This tool generates reports. At the moment it generates these reports:

 - Products whose NGAS IDs are not known by NGAS

In the future it should also generate these reports:

 - Discrepancies between the legacy archive and the new archive

In "summary mode", the tool produces a count of the problems. In "verbose mode", the tool produces a report of exactly
which products are missing.
"""
from abc import ABC, abstractmethod
import argparse
from functools import reduce
import math
from typing import List
import sys

import pandas
from pandas import DataFrame
from sqlalchemy.orm.exc import NoResultFound
from tqdm import tqdm

import pyat.schema
from pyat.datafinder.util import groups_of
from pyat.schema import ScienceProduct, AncillaryProduct, Project
from pyat.schema.ngasmodel import NGASFile


# These are the columns of the general DataFrame we use for looking up product files
COLUMNS = ['locator', 'ngas_id', 'ngas_cluster', 'ngas_location']

# This is the number of files to try at once
GROUP_SIZE = 10000


class ReporterMode(ABC):
    """
    Abstracts "details mode" and "summary mode".
    """
    @abstractmethod
    def product_found_missing(self, locator: str, found: List[str], missing: List[str]):
        """
        Report on each locator's found or missing files
        :param locator:   the locator under which we have these files
        :param found:     the files that were found
        :param missing:   the files that were missing
        :return:
        """
        pass

    @abstractmethod
    def final_report(self) -> bool:
        """
        Produce the final report.
        :return: true if all the files are present; false if some are missing.
        """
        pass


class SummaryReporterMode(ReporterMode):
    """
    Summary mode just produces a final report of how many products were found or missing.
    """
    def __init__(self):
        self.products, self.found, self.missing = 0, 0, 0

    def product_found_missing(self, locator, found, missing):
        self.products += 1
        self.found += len(found)
        self.missing += len(missing)

    def final_report(self):
        print(f'{self.products} products\n{self.found} files found\n{self.missing} files missing')
        return self.missing == 0


class DetailsReporterMode(ReporterMode):
    """
    Details mode tells you something about each missing file.
    If verbose is True, it also tells you something about every product that is not missing files.
    """
    def __init__(self):
        self.summary = SummaryReporterMode()
        self.verbose = False

    def product_found_missing(self, locator, found, missing):
        # defer to our delegate for adding things up
        self.summary.product_found_missing(locator, found, missing)

        # now print some extra stuff
        if len(missing) > 0:
            print(f'{locator} is missing {len(missing)} files')
            if self.verbose:
                for file in found:
                    print(f'  {file}')
        elif self.verbose:
            print(f'{locator} is intact')

    def final_report(self):
        return self.summary.final_report()


class ProductDegrouper(ReporterMode):
    """
    Since we chunk up the output by some number of results at a time, all of the files found and missing
    per product are not going to be sent at the same time. This "mode" just wraps another mode and
    groups together all the notices for each file by waiting for the locator to change.
    """
    def __init__(self, underlying: ReporterMode):
        self.underlying = underlying
        self.last_locator = None
        self.last_found = []
        self.last_missing = []

    def product_found_missing(self, locator: str, found: List[str], missing: List[str]):
        # new call for the previous locator? add things up
        if self.last_locator is not None and locator == self.last_locator:
            self.last_found.extend(found)
            self.last_missing.extend(missing)
        else:
            # report the previous batch
            self.underlying.product_found_missing(self.last_locator, self.last_found, self.last_missing)

            # reset things
            self.last_locator = locator
            self.last_found = found
            self.last_missing = missing

    def final_report(self):
        # send out the last batch
        self.product_found_missing(None, [], [])
        # now propagate the final report
        return self.underlying.final_report()


def find_science_product_files(session):
    """
    Find all science products.
    """
    query = """
    with recursive
         rgroups as (
            select sp.filegroup_id as top_fg, sp.filegroup_id as child_fg
            from science_products sp
            union all
            select rgroups.top_fg as top_fg, fg.filegroup_id
            from rgroups
            join filegroups fg on rgroups.child_fg = fg.parent_filegroup_id)
        select sp.science_product_locator as locator, f.ngas_id, f.ngas_cluster, f.ngas_location from science_products sp
        join rgroups on sp.filegroup_id = rgroups.top_fg
        join files f on f.filegroup = rgroups.child_fg"""
    return pandas.read_sql(query, session.connection())


def find_ancillary_product_files(session):
    """
    Find all ancillary products.
    """
    query = """
    with recursive
         rgroups as (
            select ap.filegroup_id as top_fg, ap.filegroup_id as child_fg
            from ancillary_products ap
            union all
            select rgroups.top_fg as top_fg, fg.filegroup_id
            from rgroups
            join filegroups fg on rgroups.child_fg = fg.parent_filegroup_id)
        select ap.ancillary_product_locator as locator, f.ngas_id, f.ngas_cluster, f.ngas_location from ancillary_products ap
        join rgroups on ap.filegroup_id = rgroups.top_fg
        join files f on f.filegroup = rgroups.child_fg"""
    return pandas.read_sql(query, session.connection())


def find_all_products(session):
    """
    Find all products, period
    """
    return find_science_product_files(session).append(find_ancillary_product_files(session))


def find_product_files(product_locator):
    """
    Find all the files under this product locator
    """
    def find_files(session):
        try:
            # technically this kind of comparison should only happen in the locator service, but here we are
            if 'ancillary_product' in product_locator:
                product = session.query(AncillaryProduct)\
                    .filter(AncillaryProduct.ancillary_product_locator == product_locator).one()
                return get_ancillary_product_files(product)
            else:
                product = session.query(ScienceProduct)\
                    .filter(ScienceProduct.science_product_locator == product_locator).one()
                return get_science_product_files(product)
        except NoResultFound:
            print(f'Unable to locate product {product_locator}', file=sys.stderr)
            return find_nothing(session)

    return find_files


def get_science_product_files(product: ScienceProduct) -> DataFrame:
    """
    Find all the files under this science product
    """
    result = DataFrame(((product.locator, f.ngas_id, f.ngas_cluster, f.ngas_location)
                        for f in product.filegroup.all_files),
                       columns=COLUMNS)
    for ap in product.ancillary_products:
        result = result.append(get_ancillary_product_files(ap))
    return result


def get_ancillary_product_files(product: AncillaryProduct) -> DataFrame:
    """
    Find all of the files under this ancillary product
    """
    return DataFrame(((product.locator, f.ngas_id, f.ngas_cluster, f.ngas_location)
                      for f in product.filegroup.all_files),
                     columns=COLUMNS)


def find_project_files(project_code):
    """
    Find all the files under this project
    """
    def find_files(session):
        try:
            project = session.query(Project).filter(Project.project_code == project_code).one()
            result = DataFrame([], columns=COLUMNS)
            for sp in project.science_products:
                result = result.append(get_science_product_files(sp))
            return result
        except NoResultFound:
            print(f'Unable to locate project {project_code}', file=sys.stderr)
            return find_nothing(session)
    return find_files


def combine_finders(finder1, finder2):
    """
    Combine two finder functions together into a single one
    """
    def find_products(session):
        return finder1(session).append(finder2(session))
    return find_products


def find_nothing(session):
    """
    Find nothing. This is needed as a base case for the recursion of joining together a list of finders.
    """
    return DataFrame([], columns=COLUMNS)


class DataFinder:
    def __init__(self):
        self.reporter = SummaryReporterMode()
        self.product_finder = find_all_products
        self.quiet = False
        self.verbose = False

    def main(self):
        # parse the command line
        self.parse_commandline()

        # run the process
        return self.process()

    def parse_commandline(self):
        parser = argparse.ArgumentParser(description='Find missing data in the archive')
        parser.add_argument('--science', dest='product_finder', action='append', const=find_science_product_files,
                            help='check all science products', nargs='?', default=[])
        parser.add_argument('--ancillary', dest='product_finder', action='append', const=find_ancillary_product_files,
                            help='check all ancillary products', nargs='?', default=[])
        parser.add_argument('--all', dest='product_finder', action='append', const=find_all_products,
                            help='check all products', nargs='?', default=[])
        parser.add_argument('--project', dest='projects', action='append', default=[],
                            help='check all products associated with this project')
        parser.add_argument('--product', dest='products', action='append', default=[],
                            help='check this product specifically')
        parser.add_argument('-s', dest='reporter', action='store_const', const=SummaryReporterMode(),
                            default=SummaryReporterMode(),
                            help='summary mode (just numbers please)')
        parser.add_argument('-d', dest='reporter', action='store_const', const=DetailsReporterMode(),
                            help='details mode (report for every product)')
        parser.add_argument('-q', dest='quiet', action='store_true', default=False, help='suppress progress bar')
        parser.add_argument('-v', dest='verbose', action='store_true', default=False, help='show intact files')

        args = parser.parse_args()

        # now let's get some things from the namespace
        self.quiet = args.quiet
        self.build_reporter(args)
        self.build_file_finder(args)

    def build_file_finder(self, args):
        # We have two levels of abstraction here. the lower level one is that we have a few functions that,
        # when given a SQL Alchemy session, returns a DataFrame with COLUMNS for columns. These functions are
        # find_science_product_files and find_ancillary_product_files, and find_all_product_files.
        #
        # These can be composed with each other by calling one, calling the second, and appending the output of the
        # second to the first, which is exactly what combine_finders does. You can generalize this to a list of finders
        # by using reduce, which calls combine_finders on each pair until there is just one. find_nothing acts as a
        # default if none are supplied.
        #
        # The higher layer of abstraction is the find_product_files and find_project_files functions. These need to be
        # composed with the functions discussed above, so they can only take the SQL Alchemy session as a parameter,
        # but they also need to know the name of their product or project. The solution is to make these into functions
        # that return functions of the above type, so find_product_files('foo') returns a function find_product(session)
        # that takes a session and returns the files it finds for 'foo'.
        #
        # This method first converts the higher-level finders for specific projects and products into their reduced
        # form, and then composes those with whatever low-level finders exist.

        # prepare the specific products finder
        # so the idea here is first to create a finder that gets each specific products that were asked for
        single_product_finders = reduce(combine_finders, [find_product_files(p) for p in args.products], find_nothing)

        # do the same thing, for projects
        project_product_finders = reduce(combine_finders, [find_project_files(p) for p in args.projects],
                                         find_nothing)

        # now we add that to the list of product finders from the arguments
        # that list contains finders for finding all science, all ancillary, etc.
        args.product_finder.insert(0, project_product_finders)
        args.product_finder.insert(0, single_product_finders)

        # now we combine all of the set finders into a single function that gets everything for us
        self.product_finder = reduce(combine_finders, args.product_finder, find_nothing)

    def build_reporter(self, args):
        # the reporter is a bit gross, because we have the instance above,
        # but we need to propagate the verbose setting into it
        self.reporter = args.reporter
        self.reporter.verbose = args.verbose  # yes, this is gross
        # and now we have to degroupify the file results, which we can do with a small wrapper here
        self.reporter = ProductDegrouper(self.reporter)

    def process(self):
        # step 1: open a database and get all the products
        session = pyat.schema.create_session('SDM')

        # step 2: get the products
        files = self.product_finder(session)

        # step 3: group things up so that we don't overly stress postgres remotely or memory locally
        ngas = pyat.schema.create_session('NGAS')
        grouping = self.create_groupings(files)

        # repeat by each group:
        for chunk in grouping:
            # get the NGAS IDs for this group
            ngas_ids = pandas.read_sql(ngas.query(NGASFile.file_id)
                                       .filter(NGASFile.file_id.in_(chunk.ngas_id.values)).statement,
                                       ngas.connection())

            # they are here, we just found them
            ngas_ids['present'] = True

            # join up using Pandas
            joined = ngas_ids.set_index('file_id').join(chunk.set_index('ngas_id'), how='right')

            # look through the joined up stuff, grouping it by locator per the mode API
            for spl, group in joined.groupby('locator'):
                self.reporter.product_found_missing(spl,
                                                    list(group[group.present.notna()].index),   # found
                                                    list(group[group.present.isna()].index))  # missing
        return self.reporter.final_report()

    def create_groupings(self, files):
        """
        Possibly wrap the argument iterator with a tqdm progress bar, if warranted
        :param files:
        :return:
        """
        # this is the main trick, making the grouping and knowing how many chunks there are
        grouping = groups_of(files, GROUP_SIZE)
        group_count = math.ceil(len(files) / GROUP_SIZE)

        # make the wrapping, if we're not in quiet mode and going to have more than ten chunks
        # (otherwise, the progress bar hardly shows)
        if not self.quiet and group_count >= 10:
            grouping = tqdm(grouping, total=group_count)

        return grouping


def main():
    x = DataFinder()
    sys.exit(0 if x.main() else 1)
