from __future__ import print_function

import argparse as ap
import json
import re
import sys

import dateutil.parser
import pika
from blessings import Terminal
from pycapo import CapoConfig
from pytz import timezone

from pyat import version

# Description of this widget.
_DESCRIPTION = """Command line tool for dumping out AMQP logs for the AAT/PPI, version {}."""

# Printed at the end of the help for the tool, describes return values.
_EPILOG = """Return values:
1: can't deduce CAPO_PROFILE value
2: missing CAPO property
"""

# Error messages.
_MISSING_PROFILE = """ERROR: dumplogs can't deduce the 'profile', give it the -P argument or set the CAPO_PROFILE environment variable! Geeze!\n\n"""
_MISSING_PROPERTY = "ERROR: missing required CAPO property {}\n\n"


t = Terminal()


def get_parser():
    """Build and return an argument parser for this tool."""
    p = ap.ArgumentParser(description=_DESCRIPTION.format(version),
                          formatter_class=ap.RawTextHelpFormatter,
                          epilog=_EPILOG)
    p.add_argument('--outfile', '-O', action='store',
                   help='write output to file, - for STDOUT')
    p.add_argument('-P', '--profile', action='store',
                   help='profile name to use, e.g. test, production')
    p.add_argument('-p', '--show-properties', action='store_true', default=False,
                   help='Show MDC properties')
    p.add_argument('-R', '--raw', action='store_true', help='Raw mode. No pretty formatting!')
    p.add_argument('--exclude', action='store', help='Exclude logs matching this regular expression, e.g. ".*DataFetcher.*"')

    return p


def die_message(message, parser, return_value):
    """Given an error message, argument parser and return_value, print the
    message, dump the parser's help and return with the return value."""
    sys.stderr.write(message)
    parser.print_help()
    sys.exit(return_value)


def get_settings():
    """Get all the settings this widget needs to run, some command line args,
    some are CAPO properties: returns a tuple (hostname, username, password,
    outfile)."""
    parser = get_parser()
    args = parser.parse_args()

    try:
        config = CapoConfig(profile=args.profile).settings('edu.nrao.archive.configuration.AmqpServer')

        args.hostname = config.hostname
        args.username = config.username
        args.password = config.password
    except ValueError:
        die_message(_MISSING_PROFILE, parser, 1)
    except KeyError as e:
        die_message(_MISSING_PROPERTY.format(e.args[0]), parser, 2)

    return args

class LogDumper:
    def __init__(self, settings):
        self.settings = settings
        self.credentials = pika.PlainCredentials(settings.username, settings.password)
        self.connection = pika.BlockingConnection(
            parameters=pika.ConnectionParameters(host=settings.hostname,
                                                 credentials=self.credentials))
        self.channel = self.connection.channel()
        self.queue = self.channel.queue_declare('', exclusive=True, auto_delete=True,
                                                durable=False).method.queue
        self.channel.queue_bind(self.queue, 'archive.logs')
        callback = self.raw_callback if self.settings.raw else self.callback
        self.channel.basic_consume(self.queue, on_message_callback=callback, auto_ack=True)
        if self.settings.outfile and self.settings.outfile != '-':
            self.fh = open(self.settings.outfile, 'a')
        else:
            self.fh = sys.stdout


    def raw_callback(self, ch, method, properties, body):
        parsed = json.loads(body.decode('UTF-8'))
        self.fh.write('{}: {}\n'.format(parsed['class'], parsed['formattedMessage']))


    def callback(self, ch, method, properties, body):

        try:
            parsed = json.loads(body.decode('UTF-8'))
            parsed['hostname'] = parsed['properties']['HOSTNAME'] \
                if 'HOSTNAME' in parsed['properties'] else ''
            parsed['idValue'] = parsed['properties']['track_id'] \
                if 'track_id' in parsed['properties'] else ''

            if 'formattedMessage' in parsed:
                parsed['date'] = dateutil.parser.parse(parsed['timestamp'])

                self.write(parsed)

                if 'stackTrace' in parsed:
                    # self.fh.write("\n\n ***** BEGIN STACK TRACE ******\n")
                    for stack in parsed['stackTrace']:
                        for line in stack:
                            self.fh.write('{}\n'.format(line))
                    # self.fh.write("\n\n ***** END STACK TRACE ******\n\n")
            else:
                self.fh.write('no formatted message: {}\n'.format(body))

        except Exception as e:
            self.fh.write('{}\n'.format(e))
            self.fh.write('unparseable: {}\n'.format(body))


    def dump(self):
        self.channel.start_consuming()


    def write(self, parsed):
        # self.fh.write('{} {} {} {}'.format(LEVEL['DEBUG'], LEVEL['INFO'], LEVEL['WARN'], LEVEL['ERROR']))
        parsed['hostname'] = parsed['properties']['user'] + '@' + parsed['hostname'] if 'user' in parsed['properties'] else parsed['hostname']
        parsed['hostname'] = t.blue(parsed['hostname'])
        parsed['loggerName'] = t.cyan('{0}#{1}:{2}'
                                       .format(declasse(parsed['loggerName']),
                                               parsed['method'], parsed['line']))
        parsed['level'] = LEVEL[parsed['level']]
        parsed['date'] = t.magenta(parsed['date'].astimezone(timezone('MST7MDT')).strftime('%m/%d %X.%f')[:-3])
        parsed['idValue'] = t.magenta(parsed['idValue'])
        parsed['properties'] = t.yellow_on_blue(str(parsed['properties'])) if self.settings.show_properties else ''

        log = "{hostname: <40} - {date} {level: <21} {loggerName} - {idValue} - {formattedMessage} {properties}\n".format(**parsed)
        try:
            if not re.match(self.settings.exclude, log):
                self.fh.write(log)
            else:
                log = "{hostname: <40} - {date} >>> EXCLUDED: {level: <21} {loggerName} - {formattedMessage} {properties}\n".format(**parsed)
                self.fh.write(log)
        except:
            self.fh.write(log)


LEVEL = {
    'DEBUG': t.bright_white_on_yellow('DEBUG'),
    'INFO' : t.bright_white_on_blue  ('INFO'),
    'WARN' : t.yellow_on_black       ('WARN'),
    'ERROR': t.bright_white_on_red   ('ERROR')
}


def declasse(logger):
    depath = re.sub('([a-z])[^/]+(?=/)', '\\1', logger)
    return re.sub('([a-z])[a-z]+(?=\.)', '\\1', depath)


def dumplogs():
    args = get_settings()
    dumper = LogDumper(args)
    dumper.dump()


if __name__ == '__main__':
    dumplogs()
