# -*- coding: utf-8 -*-

import argparse as ap
import collections
import datetime
import json
import logging, logging.handlers
import os
import sys
import time
from pathlib import Path

import dateutil.parser
import pendulum
from pid import PidFile, PidFileError
from pycapo import CapoConfig
from sqlalchemy import create_engine, or_
from sqlalchemy.orm import sessionmaker

from pyat.pymygdala import SendNRAOEvent, LogHandler
from pyat.mr_books import version
from pyat.schema.model import ExecutionBlock
from pyat.schema.ngasmodel import NGASFile, NGASDisk, NGASHost

_STATE_DIR = os.path.abspath(os.path.join(Path.home(), '.mr_books'))
_CACHE_FILE = 'cache.json'
_TIMESTAMP_FORMAT = '%Y-%m-%dT%H:%M:%S.%f'
_EPOCH = dateutil.parser.parse('1970-01-01T00:00.000')

_DESCRIPTION = """AAT/PPI calibration bookkeper, version {}."""

_EPILOG = """Return values:
0: everything worked,
1: can't deduce which profile to use,
2: unknown required property,
3: missing CAPO property,
4: permissions issue of some kind"""

_MISSING_PROFILE = """ERROR: mr_books can't deduce the 'profile', give it the -P argument or set the CAPO_PROFILE environment variable! Geeze!\n\n"""

_MISSING_SETTING = """ERROR: missing CAPO property {}.\n"""

_UNKNOWN_PROPERTY = """ERROR: unknown required CAPO property {}.\n"""

_RUN_STATE = 'edu.nrao.archive.workflow.config.CiplWorkflowSettings.ciplRunState'

_REQUIRED_PROPERTIES = [
    'ngasDatabase.jdbcUrl',
    'ngasDatabase.jdbcUsername',
    'ngasDatabase.jdbcPassword',
    'metadataDatabase.jdbcUrl',
    'metadataDatabase.jdbcUsername',
    'metadataDatabase.jdbcPassword',
    _RUN_STATE
]

# the magic name of a placeholder BDF
_STUB_BDFS = ['uid:__evla_bdf_X1.bdf', 'uid___evla_bdf_X1.bdf']

_NOW = time.time()

# set up logging:
LOG = logging.getLogger("mr_books")
LOG.setLevel(logging.WARN)

# Trim down the external logs we don't care about
logging.getLogger('pika').setLevel(logging.WARN)
logging.getLogger('PidFile').setLevel(logging.WARN)


def jdbc_to_python(jdbcurl, username, password):
    """ does not work for oracle """
    return jdbcurl.replace('jdbc:', '').replace('://', '://' + username + ':' + password + '@')


def get_rawdata_filegroup(eb):
    """ Return an EB's rawdata filegroup, if it can't find it throw a LookupError. """

    #
    # VLBA doesn't have the same structure, so we
    # refer to the main filegroup.
    #
    if eb.telescope == 'VLBA':
        return eb.filegroup

    if eb.telescope == 'EVLA' or eb.telescope == 'ALMA':
        for fg in eb.filegroup.children_filegroups:
            if fg.type == 'rawdata':
                return fg

    # For now, we default with an exception to catch attention
    raise LookupError("No such filegroup 'rawdata'")


def format_dt(t):
    s = t.strftime(_TIMESTAMP_FORMAT)
    f = round(float(s[-7:]), 3)
    temp = "%.3f" % f
    temp = "%s%s" % (s[:-7], temp[1:])
    return temp


class CapoSettings():
    r""" A class that pulls out the CAPO properties we care about and fails with reasonable error messages if the
    property we asked for doesn't exist. """

    def __init__(self, parser, args):
        self._parser = parser
        self._args = args
        self.profile = self._get_profile()
        self._config = self._get_capo_config()
        self.properties = dict()
        for s in _REQUIRED_PROPERTIES:
            p = self._get_capo_property(s, self._config.getstring)
            self.properties[s] = p

    def get(self, name):
        r""" Get a property by short name (the full name minus the prefix). """
        return self.properties[name].replace('\"', '')

    def _get_profile(self):
        result = None
        if self._args.profile:
            result = self._args.profile
        elif 'CAPO_PROFILE' in os.environ:
            result = os.environ['CAPO_PROFILE']
        else:
            LOG.error(_MISSING_PROFILE)
            self._parser.print_help()
            sys.exit(1)
        return result

    def _get_capo_config(self):
        r"""" Get a CapoConfig object. """
        try:
            result = CapoConfig(profile=self._args.profile)
        except ValueError:
            LOG.error(_MISSING_PROFILE)
            self._parser.print_help()
            sys.exit(1)
        return result

    def _get_capo_property(self, name, fn):
        r""" Get a CAPO property with a given short name, using a given function. """
        try:
            result = fn(name)
        except KeyError:
            LOG.error(_MISSING_SETTING.format(name))
            self._parser.print_help()
            sys.exit(2)
        return result


class NGASFileInfoCache():
    def __init__(self, capo_settings, parser, args):
        self._parser = parser
        self._args = args
        self._cs = capo_settings
        self.ngasdb_session = self._get_ngasdb_session()
        self.last_update = '2000-01-01T00:00.000'
        self._state_dir = args.state_dir
        self.cache_file = os.path.join(self._state_dir, _CACHE_FILE)
        self.cache = dict()

    def _get_ngasdb_session(self):
        try:
            # Old Oracle Connection setup
            # ngas_dsn = cx_Oracle.makedsn('quigon.aoc.nrao.edu', '1521', 'nraongas.nrao.edu')
            # ngas_dsn = ngas_dsn.replace('SID', 'SERVICE_NAME')
            # ngas_url = 'oracle+cx_oracle://ngas:XXXXXXXX@' + ngas_dsn
            # return sessionmaker(bind=create_engine(ngas_url))()
            #
            #Get the NGAS credentials from capo.
            ngas_url = jdbc_to_python(self._cs.get('ngasDatabase.jdbcUrl'),
                                      self._cs.get('ngasDatabase.jdbcUsername'),
                                      self._cs.get('ngasDatabase.jdbcPassword'))
            return sessionmaker(bind=create_engine(ngas_url))()
        except:
            LOG.exception('Exception while connecting to NGAS!')
            sys.exit(2)

    def read_cache_file(self):
        LOG.debug('reading cache file')
        result = dict()
        try:
            with open(self.cache_file, 'r') as f:
                result = json.load(f)
            dt = os.path.getmtime(self.cache_file)
            dt = datetime.datetime.utcfromtimestamp(dt)
            dt = dt - datetime.timedelta(days=2)
            self.last_update = dt
        except:
            self.last_update = _EPOCH
        return result

    def write_cache_file(self):
        LOG.debug('writing cache file')
        try:
            if not os.path.exists(self._state_dir):
                os.makedirs(self._state_dir)
            with open(self.cache_file, 'w') as f:
                json.dump(self.cache, f)
        except:
            LOG.exception('caught something while writing')
            sys.exit(2)

    def get_fileinfo_from_database(self):
        LOG.info('Updating file info from NGAS database')
        query = self.ngasdb_session.query(NGASFile).join(NGASFile.disk).join(NGASDisk.host)
        query = query.filter(or_(NGASFile.file_id.like('%.sdm'),
                                 NGASFile.file_id.like('%.bin'),
                                 NGASFile.file_id.like('%.bdf')))
        query = query.filter(NGASDisk.mounted != 0)
        query = query.filter(NGASHost.domain == 'aoc.nrao.edu')
        query = query.filter(NGASFile.ingestion_date >
                             format_dt(self.last_update))
        query = query.order_by(NGASFile.version.desc())

        result = dict()
        if query.count():
            for q in query.all():
                if q.file_id not in result:
                    result[q.file_id] = 1
        return result

    def update(self):
        LOG.debug('starting cache update')
        self.cache = self.read_cache_file()
        LOG.info('last update {}'.format(self.last_update.strftime(_TIMESTAMP_FORMAT)))
        file_info = self.get_fileinfo_from_database()
        self.cache.update(file_info)
        self.write_cache_file()

    def exists(self, file_id):
        return file_id in self.cache


class MrBooks():
    r""" MrBooks. """

    def __init__(self):
        self._parser = _make_parser()
        self._args = self._parser.parse_args()
        self._cs = CapoSettings(self._parser, self._args)
        self.profile = self._cs.profile
        self._ngas_cache = NGASFileInfoCache(self._cs, self._parser, self._args)
        self.run_state = self._cs.get(_RUN_STATE)
        self.verbose = self._args.verbose
        self._state_dir = self._args.state_dir
        self._filesets = self._args.filesets if hasattr(self._args, 'filesets') else None
        if self._filesets is not None:
            self._filesets = self._filesets.split(",")
        self.aatdb_session = self._get_aatdb_session()
        self._send_event = SendNRAOEvent(profile=self.profile, application='ingestor',
                                         routing_key='ingestion-complete.rawdata')
        try:
            if not os.path.exists(self._state_dir):
                os.makedirs(self._state_dir)
        except:
            LOG.exception('problem creating state directory')
            sys.exit(2)

    def _get_aatdb_session(self):
        try:
            aat_url = jdbc_to_python(self._cs.get('metadataDatabase.jdbcUrl'),
                                     self._cs.get('metadataDatabase.jdbcUsername'),
                                     self._cs.get('metadataDatabase.jdbcPassword'))
            return sessionmaker(bind=create_engine(aat_url))()
        except:
            LOG.exception('exception while connecting to Metadata Database')
            sys.exit(2)

    def is_calibratable(self, eb):
        r""" Is this EB one that *can* be calibrated? """
        if eb.telescope != 'EVLA':
            return False
        if eb.starttime < 56316.0:
            return False
        return True

    def should_calibrate(self, eb):
        r""" Stub function: should we calibrate this? Herein will be the heuristic that calibrates only proposed science
        and skips over tests, plus exceptions. """
        if eb.scheduling_block_type is not None and \
                (eb.scheduling_block_type == 'OBSERVER' or
                 eb.project_code.lower().startswith('vlass') or
                 eb.project_code.lower() == 'tsky0001'):
            return True
        return False

    def _unknown_ebs(self):
        r""" Returns a list of the EBs with an unknown calibration status. """
        query = self.aatdb_session.query(ExecutionBlock)
        query = query.filter(ExecutionBlock.calibration_status == 'Unknown')
        return query.all()

    def _not_ready_ebs(self):
        r""" Returns a list of the EBs with a not ready status. """
        query = self.aatdb_session.query(ExecutionBlock)
        query = query.filter(ExecutionBlock.calibration_status == 'Not Ready')
        return query.all()

    def set_initial_status(self):
        r""" Set the initial calibration status: ingestion leaves it 'Unknown', mr_books figures out whether that should
        be 'not ready' or 'do not'. Checking each 'Not Ready' is done in a different stage. """
        ebs = self._unknown_ebs()

        # For the EVLA, check for whether it is science-y (set to Not Ready) or operations-y (Set to Do Not Calibrate)
        #
        # For ALMA, assume it can be calibrated, and set it to Not Ready.
        #
        # Anything else, assume 'Do Not Calibrate'.
        #    When setting the 'Do Not Calibrate' value, send an event to trigger a re-index, since
        #    this changes the options available via the archiveIface
        #
        for eb in ebs:
            if   "EVLA" == eb.telescope:
                if self.is_calibratable(eb) and self.should_calibrate(eb):
                    # This is Observer Science (or a special case).
                    # Mark it to be tracked for ingestion completion
                    eb.calibration_status = 'Not Ready'
                else:
                    # This is unsuitable for CIPL, mark it as such and move on
                    eb.calibration_status = 'Do Not Calibrate'
                    LOG.info('{} complete: Do Not Calibrate'.format(eb.ngas_fileset_id))
                    self.fire_off_event(eb)

            elif "ALMA" == eb.telescope:
                # update to Not Ready status, the rest is handled later
                eb.calibration_status = 'Not Ready'
            else:
                #
                # GBT, VLBA, and anything else all are marked Do Not Calibrate
                # in our scheme (subject to change later).
                #
                eb.calibration_status = 'Do Not Calibrate'
                LOG.debug('{} is not EVLA/ALMA: Do Not Calibrate'.format(eb.ngas_fileset_id))
                self.fire_off_event(eb)

        # Update the execution_blocks table:
        try:
            self.aatdb_session.add_all(ebs)
            self.aatdb_session.commit()
        except:
            self.aatdb_session.rollback()
            sys.exit(5)

    def _get_eb(self, fileset_id):
        query = self.aatdb_session.query(ExecutionBlock)
        query = query.filter(ExecutionBlock.ngas_fileset_id == fileset_id)
        return query.first()

    def get_ebs(self, fileset_ids):
        r""" Get an EB by a specific ebs: this is mostly for testing right now. """
        result = list()
        if isinstance(fileset_ids, collections.Iterable):
            for fileset_id in fileset_ids:
                result.append(self._get_eb(fileset_id))
        else:
            result.append(self._get_eb(fileset_ids))
        return result

    def is_eb_complete(self, eb):
        r""" Check to see if an EB is complete, meaning all the files in the filegroup are in NGAS. """

        files = get_rawdata_filegroup(eb).files
        LOG.debug('checking {}, {} files'.format(eb.ngas_fileset_id, len(files)))
        found = 0
        for f in files:
            # _STUB_BDF is a magic BDF name, a placeholder: don't bother checking to make sure it exists
            if f.ngas_id not in _STUB_BDFS:
                if not self._ngas_cache.exists(f.ngas_id):
                    LOG.debug('{} is missing'.format(f.ngas_id))
                else:
                    found += 1
        if found != len(files):
            LOG.info('{} not complete, found {} of {} files'.format(eb.ngas_fileset_id, found, len(files)))
            return False
        else:
            LOG.info('{} complete, found {} of {} files'.format(eb.ngas_fileset_id, found, len(files)))
            return True

    def fire_off_event(self, eb):
        r""" Fire off an ingestion complete event for this EB. """
        LOG.debug('firing off event')
        try:
            event = self.ingestion_event(eb)
            LOG.debug(json.dumps(event))
            self._send_event.send(event=event)
        except:
            LOG.exception('Exception occurred while sending completion event')
            sys.exit(5)
        LOG.debug('done')

    def update_to_ready(self, e):
        r""" Mark an EB (or list of EBs) as ready to calibrate. """
        LOG.debug('updating to ready')
        if isinstance(e, collections.Iterable):
            for eb in e:
                self.update_to_ready(eb)
        else:
            try:
                e.calibration_status = 'Ready'
                self.aatdb_session.add(e)
                self.aatdb_session.commit()
            except:
                LOG.exception('Exception while updating status of {}', e.ngas_fileset_id)
                self.aatdb_session.rollback()
                sys.exit(5)
        LOG.debug('done')

    def check_not_ready(self, e):
        r""" Check an EB (or a list of EBs) to see if they are ready to calibrate. """
        if isinstance(e, collections.Iterable):
            for eb in e:
                self.check_not_ready(eb)
        else:
            LOG.debug('checking {}'.format(e.ngas_fileset_id))
            if e.telescope == 'EVLA':
                # Check to see if all the BDFs have been ingested
                if self.is_eb_complete(e):
                    LOG.debug('{} complete: Ready'.format(e.ngas_fileset_id))
                    self.update_to_ready(e)
                    self.fire_off_event(e)
                else:  # Incomplete, wait for next go around
                    LOG.debug('{} not ready'.format(e.ngas_fileset_id))
            elif e.telescope == 'ALMA':
                # ALMA is the easy case, just set it to Ready
                LOG.info('{} set to Ready automatically'.format(e.ngas_fileset_id))
                self.update_to_ready(e)
                self.fire_off_event(e)
            else:
                # Everything else, just print a message
                LOG.info('Unexpected Telescope for {}'.format(e.ngas_fileset_id))

    def ingestion_event(self, eb):
        result = dict()
        result['logData'] = {
            'profile': self.profile,
            'fileset_id': eb.ngas_fileset_id,
            'ingestion_type': 'observation',
            'telescope': eb.telescope,
            'execblock_id': eb.ost_exec_block_id,
            'schedblock_id': eb.scheduling_block_id,
            'schedblock_type': eb.scheduling_block_type,
            'project_code': eb.project_code,
            'override_db_screen': 'false',
            'file_count': len(get_rawdata_filegroup(eb).files)
        }
        result['message'] = 'ingestion complete'
        result['request'] = 'de nada'
        return result

    def update_to_dnc(self, e):
        r""" Update an EB (or list of EBs) to Do Not Calibrate. """
        if isinstance(e, collections.Iterable):
            for eb in e:
                self.update_to_dnc(eb)
        else:
            try:
                e.calibration_status = 'Do Not Calibrate'
                self.aatdb_session.add(e)
                self.aatdb_session.commit()
            except:
                LOG.exception('Exception while updating calibration status of {}', e.ngas_fileset_id)
                self.aatdb_session.rollback()
                sys.exit(5)

    def set_ingestion_complete(self):
        pass

    def run(self):
        r""" Run the actual clean process. """
        start_time = pendulum.now()

        try:
            with PidFile(piddir=self._state_dir, pidname='mr_books.lock') as p:
                self._ngas_cache.update()

                LOG.debug('mr_books called with arguments:\n{}'.format(self._args))

                if self._filesets:
                    ebs = self.get_ebs(self._filesets)
                    self.check_not_ready(ebs)
                else:
                    self.set_ingestion_complete()
                    self.set_initial_status()
                    self.check_not_ready(self._not_ready_ebs())
                    time_diff = pendulum.now() - start_time
                    LOG.info('finished, took {}'.format(time_diff.in_words()))
        except PidFileError:
            LOG.error("mr_books already running, aborting this instance")


def _make_parser():
    r""" Build a command line parser for this app: this is external to MrBooks and/or CapoSettings because both need it,
    and Sphinx will want a look at ot to build docs. """
    result = ap.ArgumentParser(description=_DESCRIPTION.format(version),
                               formatter_class=ap.RawTextHelpFormatter,
                               epilog=_EPILOG)
    result.add_argument('-P', '--profile', action='store',
                        help='profile name to use, e.g. test, production')
    result.add_argument('-v', '--verbose', action='store_true',
                        help='enable verbose-i-tude')
    result.add_argument('-f', '--filesets', action='store', dest='filesets',
                        help='run on comma separated list of filesets')
    result.add_argument('-s', '--state_dir', action='store', dest='state_dir',
                        default=_STATE_DIR,
                        help='state directory to use ({})'.format(_STATE_DIR))
    return result


def mr_books():
    r""" Wrapper around the MrBooks class, console script entry point. """


    mb = MrBooks()

    # Configure the logging:
    # Log specifically to a file in the state directory,
    # rather than to stdout.  Rotate weekly, on Friday.
    logFileName = mb.profile + '_direct.log'
    fullLogPath = os.path.abspath( os.path.join(mb._state_dir, logFileName) )
    ch = logging.handlers.TimedRotatingFileHandler(fullLogPath, when='W4')
    if mb.verbose:
        # Given the the -v flag, be extra chatty
        ch.setLevel(logging.DEBUG)
        # and open up the limits for most of our subsidiaries
        LOG.setLevel(logging.DEBUG)
    else:
        # default to WARN level to avoid some spam.
        ch.setLevel(logging.INFO)
        LOG.setLevel(logging.INFO)
    formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
    ch.setFormatter(formatter)
    LOG.addHandler(ch)

    # integrate the logs with pymygdala's logging system
    handler = LogHandler(profile=mb.profile, application='mr_books')
    if mb.verbose:
        # Given the the -v flag, be extra chatty
        handler.setLevel(logging.DEBUG)
    else:
        # default to WARN level to avoid some spam.
        handler.setLevel(logging.WARN)
    LOG.addHandler(handler)

    LOG.debug('starting, profile is {}'.format(mb.profile))

    mb.run()


if __name__ == "__main__":
    mr_books()
