# coding: utf-8
from sqlalchemy import ARRAY, BigInteger, Boolean, CheckConstraint, Column, Date, DateTime, Float, ForeignKey, Integer, JSON, LargeBinary, Numeric, String, Table, Text, text
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declarative_base


Base = declarative_base()
metadata = Base.metadata


class Author(Base):
    __tablename__ = 'authors'

    author_id = Column(Integer, primary_key=True, server_default=text("nextval('authors_author_id_seq'::regclass)"))
    project_code = Column(ForeignKey('projects.project_code'), nullable=False)
    username = Column(String, nullable=False)
    firstname = Column(String, nullable=False)
    lastname = Column(String, nullable=False)
    pst_person_id = Column(String)
    is_pi = Column(Boolean, nullable=False)

    project = relationship('Project')

    def __repr__(self):
        return "<Author#{author_id} {username} '{firstname} {lastname}'>"\
            .format(**self.__dict__)


class CalibrationStatusValue(Base):
    __tablename__ = 'calibration_status_values'

    status = Column(String, primary_key=True)
    description = Column(String, nullable=False)


    def __repr__(self):
        return "<CalibrationStatusValue {status}>".format(**self.__dict__)


class CalibrationTable(Base):
    __tablename__ = 'calibration_tables'

    calibration_table_id = Column(Integer, primary_key=True, server_default=text("nextval('calibration_tables_calibration_table_id_seq'::regclass)"))
    file_id = Column(ForeignKey('files.file_id'), nullable=False)
    metadata_tbd = Column(String, nullable=False)

    file = relationship('File')

    def __repr__(self):
        return "<CalibrationTable#{calibration_table_id}>".format(**self.__dict__)


class DataDescription(Base):
    __tablename__ = 'data_descriptions'

    data_description_id = Column(Integer, primary_key=True, server_default=text("nextval('data_descriptions_data_description_id_seq'::regclass)"))
    bandwidth = Column(Float(53), nullable=False)
    frequency = Column(Float(53), nullable=False)
    polarization_id = Column(ForeignKey('polarizations.polarization_id'), nullable=False)

    polarization = relationship('Polarization')
    subscans = relationship('Subscan', secondary='subscan_data_descriptions')

    def __repr__(self):
        return "<DataDescription#{data_description_id} Freq={frequency} BW={bandwidth}>"\
            .format(**self.__dict__)


t_datadesc = Table(
    'datadesc', metadata,
    Column('frequency', Float(53)),
    Column('bandwidth', Float(53)),
    Column('pol_id', Integer),
    Column('receiver', String),
    Column('datadesc_id', Integer),
    Column('spectral_window_id', Integer),
    Column('receiver_id', Integer)
)


class Event(Base):
    __tablename__ = 'events'

    id = Column(Integer, primary_key=True, server_default=text("nextval('events_id_seq'::regclass)"))
    application = Column(String, nullable=False)
    user = Column(String)
    request = Column(String)
    message = Column(String, nullable=False)
    log_data = Column(JSON)
    version = Column(Float(53), nullable=False)
    timestamp = Column(DateTime(True), nullable=False)

    def __repr__(self):
        return "<Event#{id} app={application} user={user} request={request} timestamp={timestamp}>"\
            .format(**self.__dict__)


t_execblock_start_stop = Table(
    'execblock_start_stop', metadata,
    Column('execution_block_id', Integer),
    Column('starttime', Float(53)),
    Column('endtime', Float(53))
)


class ExecutionBlock(Base):
    __tablename__ = 'execution_blocks'

    execution_block_id = Column(Integer, primary_key=True, server_default=text("nextval('execution_blocks_execution_block_id_seq'::regclass)"))
    ost_exec_block_id = Column(Integer)
    filegroup_id = Column(ForeignKey('filegroups.filegroup_id'), nullable=False)
    calibration_level = Column(String, nullable=False)
    telescope = Column(String, nullable=False)
    configuration = Column(String, nullable=False)
    ingestion_complete = Column(Boolean, nullable=False)
    scheduling_block_id = Column(Integer)
    ngas_fileset_id = Column(String, nullable=False)
    project_code = Column(ForeignKey('projects.project_code'), nullable=False)
    starttime = Column(Float(53))
    endtime = Column(Float(53))
    calibration_status = Column(ForeignKey('calibration_status_values.status'), nullable=False, server_default=text("'Unknown'::character varying"))
    scheduling_block_type = Column(String)

    calibration_status_value = relationship('CalibrationStatusValue')
    filegroup = relationship('Filegroup')
    project = relationship('Project')
    scans = relationship('Scan')

    def __repr__(self):
        return "<ExecutionBlock#{execution_block_id} project={project_code} start={starttime} end={endtime}>".format(**self.__dict__)



class Filegroup(Base):
    __tablename__ = 'filegroups'
    __table_args__ = (
        CheckConstraint('(CASE WHEN (project_code IS NOT NULL) THEN 1 ELSE 0 END + CASE WHEN (parent_filegroup_id IS NOT NULL) THEN 1 ELSE 0 END) = 1'),
    )

    filegroup_id = Column(Integer, primary_key=True, server_default=text("nextval('filegroups_filegroup_id_seq'::regclass)"))
    project_code = Column(ForeignKey('projects.project_code'))
    groupname = Column(String, nullable=False)
    parent_filegroup_id = Column(ForeignKey('filegroups.filegroup_id'))
    datasize = Column(BigInteger)

    parent_filegroup = relationship('Filegroup', remote_side=[filegroup_id], backref='children_filegroups')
    project = relationship('Project')

    execution_blocks = relationship('ExecutionBlock')
    files = relationship('File')
    scans = relationship('Scan')

    def __repr__(self):
        return "<Filegroup#{filegroup_id} project={project_code} groupname={groupname}>".format(**self.__dict__)


class File(Base):
    __tablename__ = 'files'

    file_id = Column(Integer, primary_key=True, server_default=text("nextval('files_file_id_seq'::regclass)"))
    file_path = Column(String)
    ngas_id = Column(String)
    filegroup = Column(ForeignKey('filegroups.filegroup_id'), nullable=False, index=True)
    filename = Column(String, nullable=False)
    filesize = Column(BigInteger, nullable=False)
    format = Column(String, nullable=False)
    type = Column(String, nullable=False)
    checksum = Column(String)
    checksum_type = Column(String)
    ingestion_time = Column(DateTime(True))

    filegroup1 = relationship('Filegroup')
    workflows = relationship('Workflow', secondary='workflow_input_files')
    calibration_tables = relationship('CalibrationTable')
    images = relationship('Image')
    subscans = relationship('Subscan')

    def __repr__(self):
        return "<File#{file_id} {file_path}/{filename} ngas_id={ngas_id}>".format(**self.__dict__)


t_flatfiles = Table(
    'flatfiles', metadata,
    Column('project_code', String),
    Column('telescope', String),
    Column('config', String),
    Column('sb_id', BigInteger),
    Column('eb_id', BigInteger),
    Column('calibration', String),
    Column('file_path', String),
    Column('ngas_file_id', String),
    Column('ngas_fileset_id', String),
    Column('filename', String),
    Column('filesize', BigInteger),
    Column('format', String),
    Column('type', String),
    Column('starttime', Numeric),
    Column('endtime', Numeric),
    Column('sourcename', String),
    Column('sourcetype', String),
    Column('ra', Numeric),
    Column('dec', Numeric),
    Column('exposure_time', Numeric),
    Column('integration_time', Numeric),
    Column('observation_type', String),
    Column('scan', Numeric),
    Column('subscan', Numeric),
    Column('intent', String),
    Column('data_desc_id', BigInteger),
    Column('generated_eb_id', BigInteger),
    Column('generated_scan_id', BigInteger)
)


class Image(Base):
    __tablename__ = 'images'

    image_id = Column(Integer, primary_key=True, server_default=text("nextval('images_image_id_seq'::regclass)"))
    file_id = Column(ForeignKey('files.file_id'), nullable=False)
    target_name = Column(String, nullable=False)
    center_position = Column(String, nullable=False)
    observing_band = Column(String, nullable=False)
    project_code = Column(String, nullable=False)
    telescope = Column(String, nullable=False)
    configuration = Column(String, nullable=False)
    collection_name = Column(String)
    thumbnail = Column(LargeBinary)
    spatial_resolution = Column(Float(53), nullable=False)
    field_of_view = Column(Float(53), nullable=False)
    max_intensity = Column(Float(53), nullable=False)
    min_intensity = Column(Float(53), nullable=False)
    rms_noise = Column(Float(53), nullable=False)
    polarization_id = Column(ForeignKey('polarizations.polarization_id'), nullable=False)

    file = relationship('File')
    polarization = relationship('Polarization')

    def __repr__(self):
        return "<Image#{image_id} target={target_name} telescope={telescope}>".format(**self.__dict__)


class Intent(Base):
    __tablename__ = 'intents'

    intent_name = Column(String, primary_key=True)

    subscans = relationship('Subscan', secondary='subscan_intents')

    def __repr__(self):
        return "<Intent {0}>".format(self.intent_name)


t_logs = Table(
    'logs', metadata,
    Column('filename', String),
    Column('class', String),
    Column('method', String),
    Column('line', Integer),
    Column('arguments', ARRAY(String)),
    Column('timestamp', DateTime(True)),
    Column('formatted_message', String),
    Column('logger_name', String),
    Column('level', String),
    Column('thread_name', String),
    Column('reference_mask', Integer),
    Column('properties', JSON),
    Column('stacktrace', Text)
)


class Polarization(Base):
    __tablename__ = 'polarizations'

    polarization_id = Column(Integer, primary_key=True)
    name = Column(String, nullable=False)
    description = Column(String, nullable=False)

    def __repr__(self):
        return "<Polarization {0}>".format(self.name)


class Project(Base):
    __tablename__ = 'projects'

    project_code = Column(String, primary_key=True)
    legacy_id = Column(String)
    total_observation_time = Column(Float(53))
    opt_project_id = Column(Integer)
    title = Column(String)
    abstract = Column(Text)
    proprietary_expiration = Column(DateTime)
    last_addition = Column(Date)
    starttime = Column(Float(53))
    endtime = Column(Float(53))
    proprietary_duration = Column(Float(53))

    authors = relationship('Author')
    execution_blocks = relationship('ExecutionBlock')
    file_groups = relationship('Filegroup')

    def __repr__(self):
        return '<Project#{project_code} "{title}" start={starttime} end={endtime}>'.format(**self.__dict__)


class Receiver(Base):
    __tablename__ = 'receivers'

    receiver_id = Column(Integer, primary_key=True, server_default=text("nextval('receivers_receiver_id_seq'::regclass)"))
    description = Column(String, nullable=False)

    def __repr__(self):
        return '<Receiver {0}>'.format(self.receiver_id)


class Scan(Base):
    __tablename__ = 'scans'

    scan_id = Column(Integer, primary_key=True, server_default=text("nextval('scans_scan_id_seq'::regclass)"))
    ost_scan_id = Column(Integer)
    execution_block_id = Column(ForeignKey('execution_blocks.execution_block_id'), nullable=False)
    filegroup_id = Column(ForeignKey('filegroups.filegroup_id'), nullable=False)
    max_bandwidth = Column(Float(53), nullable=False)
    min_bandwidth = Column(Float(53), nullable=False)
    polarization_code = Column(Integer, nullable=False)
    max_frequency = Column(Float(53), nullable=False)
    min_frequency = Column(Float(53), nullable=False)

    execution_block = relationship('ExecutionBlock')
    filegroup = relationship('Filegroup')
    subscans = relationship('Subscan')

    def __repr__(self):
        return '<Scan#{0}>'.format(self.scan_id)


t_subscan_data_descriptions = Table(
    'subscan_data_descriptions', metadata,
    Column('subscan_id', ForeignKey('subscans.subscan_id'), primary_key=True, nullable=False),
    Column('data_description_id', ForeignKey('data_descriptions.data_description_id'), primary_key=True, nullable=False, index=True)
)


t_subscan_intents = Table(
    'subscan_intents', metadata,
    Column('intent_name', ForeignKey('intents.intent_name'), primary_key=True, nullable=False),
    Column('subscan_id', ForeignKey('subscans.subscan_id'), primary_key=True, nullable=False)
)


class Subscan(Base):
    __tablename__ = 'subscans'

    subscan_id = Column(Integer, primary_key=True, server_default=text("nextval('subscans_subscan_id_seq'::regclass)"))
    scan_id = Column(ForeignKey('scans.scan_id'), nullable=False)
    file_id = Column(ForeignKey('files.file_id'), nullable=False)
    ost_subscan_id = Column(Integer)
    obstype = Column(String, nullable=False)
    starttime = Column(Float(53), nullable=False)
    endtime = Column(Float(53), nullable=False)
    sourcename = Column(String, nullable=False)
    sourcetype = Column(String, nullable=False)
    ra = Column(Float(53), nullable=False)
    dec = Column(Float(53), nullable=False)
    exposure_time = Column(Float(53), nullable=False)
    integration_time = Column(Float(53), nullable=False)
    receiver_id = Column(ForeignKey('receivers.receiver_id'), nullable=False)
    backend = Column(String, nullable=False)
    intent = Column(String)
    datadesc = Column(Integer)

    file = relationship('File')
    receiver = relationship('Receiver')
    scan = relationship('Scan')
    data_descriptions = relationship('DataDescription', secondary='subscan_data_descriptions')
    intents = relationship('Intent', secondary='subscan_intents')

    def __repr__(self):
        return '<Subscan#{subscan_id} {obstype} start={starttime} end={endtime} ra={ra} dec={dec} backend={backend} intent={intent}>'.format(**self.__dict__)


t_workflow_input_files = Table(
    'workflow_input_files', metadata,
    Column('workflow_id', ForeignKey('workflows.workflow_id'), primary_key=True, nullable=False),
    Column('file_id', ForeignKey('files.file_id'), primary_key=True, nullable=False)
)


class WorkflowParameter(Base):
    __tablename__ = 'workflow_parameters'

    workflow_id = Column(ForeignKey('workflows.workflow_id'), primary_key=True, nullable=False)
    parameter_name = Column(String, primary_key=True, nullable=False)
    parameter_value = Column(String, nullable=False)

    workflow = relationship('Workflow')

    def __repr__(self):
        return '<WorkflowParameter {parameter_name}={parameter_value}>'.format(**self.__dict__)


class Workflow(Base):
    __tablename__ = 'workflows'

    workflow_id = Column(Integer, primary_key=True, server_default=text("nextval('workflows_workflow_id_seq'::regclass)"))
    workflow_name = Column(String, nullable=False)
    status = Column(String, nullable=False)
    submitted_by = Column(String, nullable=False)
    submitted = Column(DateTime)
    started = Column(DateTime)
    completed = Column(DateTime)
    workflow_qa_analyst = Column(String)
    workflow_qa_comments = Column(String)
    workflow_qa_result = Column(String)
    result_qa_analyst = Column(String)
    result_qa_comments = Column(String)
    result_qa_results = Column(String)
    software_version = Column(String, nullable=False)
    output_filegroup_id = Column(ForeignKey('filegroups.filegroup_id'), nullable=False)

    output_filegroup = relationship('Filegroup')
    workflow_parameters = relationship('WorkflowParameter')
    input_files = relationship('File', secondary='workflow_input_files')

    def __repr__(self):
        return '<Workflow#{workflow_id} {workflow_name} status={status} start={started} end={completed}>'.format(**self.__dict__)
