# -*- coding: utf-8 -*-

from __future__ import print_function

import argparse as ap
import logging
import os
import shutil
import sys

import pendulum
import time
from pycapo import CapoConfig

from pyat import version
from pyat.pymygdala import LogHandler, SendNRAOEvent

_DESCRIPTION = """AAT/PPI request file/directory cleaner, version {}. Cleans up the spool directory and downloads area."""

_EPILOG = """Return values:
0: everything worked,
1: can't deduce which profile to use,
2: unknown required property,
3: missing CAPO property,
4: permissions issue of some kind"""

_MISSING_PROFILE = """ERROR: mr_clean can't deduce the 'profile', give it the -P argument or set the CAPO_PROFILE environment variable! Geeze!\n\n"""

_MISSING_SETTING = """ERROR: missing CAPO property {}.\n"""

_UNKNOWN_PROPERTY = """ERROR: unknown required CAPO property {}.\n"""

_REQUIRED_PROPERTIES = ['spoolDirectory', 'spoolDaysToKeep',
                        'stageDirectory', 'stageDaysToKeep',
                        'tempDirectory', 'tempDaysToKeep',
                        'downloadDirectory', 'downloadDaysToKeep',
                        'cacheDirectory', 'cacheDaysToKeep',
                        'parallelIngestionDirectory', 'parallelIngestionDaysToKeep',  # No parallel in va, so these point at tmp there
                        'enabled']

_PROPERTY_PREFIX = "edu.nrao.archive.configuration.JanitorSettings"

_NOW = time.time()

_LOG = logging.getLogger('mr_clean')
_LOG.setLevel(logging.DEBUG)
ch = logging.StreamHandler(sys.stdout)
ch.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
ch.setFormatter(formatter)
_LOG.addHandler(ch)


def _make_parser():
    r""" Build a command line parser for this app: this is external to MrClean and/or CapoSettings because both need
    it, and Sphinx will want a look at ot to build docs. """
    result = ap.ArgumentParser(description=_DESCRIPTION.format(version),
                               formatter_class=ap.RawTextHelpFormatter,
                               epilog=_EPILOG)
    result.add_argument('-P', '--profile', action='store',
                        help='profile name to use, e.g. test, production')
    result.add_argument('-v', '--verbose', action='store_true',
                        help='enable verbose-i-tude')
    return result


def contained_dirs(dir):
    r""" Get a list of the subdirectories of a given directory. """
    return filter(os.path.isdir,
                  [os.path.join(dir, f) for f in os.listdir(dir)])


def older_than(f, n):
    r""" Is a file (or directory) older than N days. """
    if os.path.islink(f):
        fn = os.lstat
    else:
        fn = os.stat
    return True if fn(f).st_mtime <= _NOW - (n * 86400) else False


class CapoSettings():
    r""" A class that pulls out the CAPO properties we care about and fails with reasonable error messages if the
    property we asked for doesn't exist. """

    def __init__(self, capo_config, parser):
        self._config = capo_config
        self._parser = parser
        self.properties = dict()

        for s in _REQUIRED_PROPERTIES:
            if s.endswith('Directory'):
                p = self._get_capo_property(_PROPERTY_PREFIX + '.' + s, self._config.getstring)
            elif s.endswith('DaysToKeep'):
                p = self._get_capo_property(_PROPERTY_PREFIX + '.' + s, self._config.getint)
            elif s == 'enabled':
                p = self._get_capo_property(_PROPERTY_PREFIX + '.' + s, self._config.getboolean)
            else:
                _LOG.error(_UNKNOWN_PROPERTY.format(_PROPERTY_PREFIX + "." + s))
                self._parser.print_help()
                sys.exit(2)
            self.properties[s] = p

    def get(self, name):
        r""" Get a property by short name (the full name minus the prefix). """
        return self.properties[name]

    def _get_capo_property(self, name, fn):
        r""" Get a CAPO property with a given short name, using a given function. """
        try:
            result = fn(name)
        except KeyError:
            _LOG.error(_MISSING_SETTING.format(name))
            self._parser.print_help()
            sys.exit(2)
        return result


class MrClean():
    r""" MrClean just keeps things clean. """

    def __init__(self):
        self._parser = _make_parser()
        self._args = self._parser.parse_args()
        self._config = self._get_capo_config()
        self._setup_logging()
        self._cs = CapoSettings(self._config, self._parser)
        self.enabled = self._cs.get('enabled')
        self.verbose = self._args.verbose
        self._send_event = SendNRAOEvent(profile=self.profile, application='janitor')

    def _setup_logging(self):
        handler = LogHandler(profile=self.profile, application='mr_clean',
                             level=logging.DEBUG)
        _LOG.addHandler(handler)

    def _get_capo_config(self):
        r"""" Get a CapoConfig object. """
        if self._args.profile:
            self.profile = self._args.profile
        elif 'CAPO_PROFILE' in os.environ:
            self.profile = os.environ['CAPO_PROFILE']
        else:
            _LOG.error(_MISSING_PROFILE)
            self._parser.print_help()
            sys.exit(1)
        return CapoConfig(profile=self.profile)

    def clean_flat_directory(self, directory, days):
        r""" Clean a 'flat' directory, this is a directory with sub-directories, we only don't descend into them, we
        only look at the modification date of each subfolder to determine whether to keep it or not. The spool folder
        and the temp folder fit this use case.  For the case of long-running cluster processes (casa in particular)
        we need to determine if there are newer subdirectories underneath the to-be-removed target."""
        if not os.path.exists(directory):
            _LOG.info('directory {} does not exist, skipping'.format(directory))
            return

        dirs = contained_dirs(directory)
        for d in dirs:
            if older_than(d, days):
                # d is a candidate for deletion, but check for newer subdirectories:
                subdirs = contained_dirs(d)
                if subdirs:
                    # non-empty set of subdirectories:
                    for sd in subdirs:
                        if not older_than(sd, days):
                            # One of the sub directories is newer, so we
                            # set the parent's modified time to now.
                            if self.verbose:
                                _LOG.info('updated modified time of directory {}'.format(d))
                            os.utime(d, None)  # The None causes the time to be set to Now
                            break
                    else:
                        # If all the subdirectories are older, remove d
                        # Note: A break above will not execute this portion
                        if self.verbose:
                            _LOG.info('removing directory {}'.format(d))
                        shutil.rmtree(d, ignore_errors=True)
                else:
                    # If there are no subdirectories, we're ok to remove d
                    if self.verbose:
                        _LOG.info('removing directory {}'.format(d))
                    shutil.rmtree(d, ignore_errors=True)

    def clean_download_directory(self, directory, days):
        r""" Clean a download directory: first step through it and nuke all the old files, then walk it back to front
        removing empty directories as we go. """
        if not os.path.exists(directory):
            _LOG.info('directory {} does not exist, skipping'.format(directory))
            return

        # Nuke old files
        for root, dirs, files in os.walk(directory, topdown=False):
            for name in files:
                fp = os.path.join(root, name)
                if older_than(fp, days):
                    if self.verbose:
                        _LOG.info('removing file {}'.format(fp))
                        os.remove(fp)

        # Nuke empty directories
        for root, dirs, files in os.walk(directory, topdown=False):
            for name in dirs:
                dp = os.path.join(root, name)
                if not os.listdir(dp):
                    if self.verbose:
                        _LOG.info('removing directory {}'.format(dp))
                        os.rmdir(dp)

    def send_event(self, event):
        self._send_event.send(event=event)

    def clean(self):
        r""" Run the actual clean process. """
        self.send_event({'message': 'starting up',
                         'request': 'cleanup',
                         'logData': {}
                         })
        start_time = pendulum.now()
        if self.verbose:
            _LOG.debug('CLI arguments:\n{}'.format(self._args))
            _LOG.debug('required CAPO properties:\n{}'.format(self._cs.properties))
        if self.enabled:
            self.clean_flat_directory(self._cs.get('spoolDirectory'), self._cs.get('spoolDaysToKeep'))
            self.clean_flat_directory(self._cs.get('tempDirectory'), self._cs.get('tempDaysToKeep'))
            self.clean_flat_directory(self._cs.get('parallelIngestionDirectory'), self._cs.get('parallelIngestionDaysToKeep'))
            self.clean_flat_directory(self._cs.get('stageDirectory'), self._cs.get('stageDaysToKeep'))

            try:
                sd = contained_dirs(self._cs.get('downloadDirectory'))
                for s in sd:
                    self.clean_download_directory(s, self._cs.get('downloadDaysToKeep'))
            except FileNotFoundError:
                _LOG.exception('caught an exception cleaning the download directory')
                _LOG.info('directory {} does not exist, skipping'.format(self._cs.get('downloadDirectory')))

            # the cache is complicated, much like downloads.  Use the same functionality:
            try:
                sd = contained_dirs(self._cs.get('cacheDirectory'))
                for s in sd:
                    self.clean_download_directory(s, self._cs.get('cacheDaysToKeep'))
            except FileNotFoundError:
                _LOG.exception('caught an exception cleaning the cache directory')
                _LOG.info('directory {} does not exist, skipping'.format(self._cs.get('cacheDirectory')))

        time_diff = pendulum.now() - start_time
        self.send_event({'message': 'finishing',
                         'request': 'cleanup',
                         'logData': {'elapsed_time': time_diff.in_words()}
                         })


def mr_clean():
    r""" Wrapper around the MrClean class, console script entry point. """
    mc = MrClean()
    mc.clean()


if __name__ == "__main__":
    mr_clean()
