# -*- coding: utf-8 -*-
"""
A module for updating the proprietary period of a project, whether that be to set it as
proprietary or to make it public.

Author: Richard Falardeau <rfalarde@nrao.edu>
"""
# pylint: disable=logging-format-interpolation

import argparse as ap
import sys
import warnings
import logging
from astropy.time import Time
from sqlalchemy import exc as sa_exc

from pyat import version, get_console_logger, LOG_MESSAGE_FORMATTER, get_my_capo_config
from pyat.schema import create_session
from pyat.schema.model import Project
from pyat.schema.legacy_model import LegacyProject
from pyat.pymygdala import LogHandler, SendNRAOEvent


_APPLICATION_NAME = 'proprietary_setter'
_LOG = get_console_logger(_APPLICATION_NAME)
_MISSING_PROFILE = """unable to derive the capo profile from the environment, """ +\
                   """provide the capo profile through the -P argument or set the """ +\
                   """CAPO_PROFILE environment variable."""
_DISALLOWED_DURATION = """invalid proprietary duration, only integer values """ +\
                       """between 0 and 730 are allowed."""
_DESCRIPTION = """A tool for setting the proprietary duration of a project"""
_EPILOG = """Return values:
0: everything worked,
1: can't deduce which profile to use
2: invalid proprietary duration
3: DB update failed"""


class UpdateException(Exception):
    r"""A wrapper for exceptions caught during our attempt to update the proprietary period.
    Since we aren't passing the argparse object to helper functions, we want to catch some
    obviously library exceptions in sqlalchemy, but also use it for some error trapping and
    'handling'"""


def _make_parser():
    r"""
    Build a command line parser to take the parameters for setting a projects
    proprietary period.

    :return result: an argparse object with the profile, project and
    duration in its namespace"""
    result = ap.ArgumentParser(description=_DESCRIPTION.format(version),
                               formatter_class=ap.RawTextHelpFormatter,
                               epilog=_EPILOG)
    result.add_argument('-P', '--profile', action='store',
                        help='profile name to use, e.g. nmtest, dsoc-test, or nmprod')
    result.add_argument('project', action='store',
                        help='project_code to update proprietary duration')
    result.add_argument('duration', action='store',
                        type=int,
                        help='an integer duration to apply to the project; '
                             '0 (immediately public) to 730 (private for two years from today)')
    return result


def set_project_proprietary_state(capo_config, project_code, proprietary_duration):
    r"""
    Set the proprietary period on the project code to the proprietary period provide on both the
    archive and the legacy archive.
        Note:   Since the capo_config will only really set the db parameters for the archive,
                all updates to the legacy archive are actually live on production.
    :param capo_config: the capo_config we're running under, which determines which db we update
    :param project_code: the project code to update
    :param proprietary_duration: an integer value for the new proprietary period (in days)
    """
    try:
        with warnings.catch_warnings():
            # This is solely to suppress the SQLAlchemy warning messages
            warnings.simplefilter("ignore", category=sa_exc.SAWarning)
            a_session = create_session('SDM', profile=capo_config.profile)
            legacy_session = create_session('LEGACY', profile=capo_config.profile)
            new_mjd_endtime = Time.now().mjd

            # In the archive, the project code is a PK, we should only ever get one
            project = a_session.query(Project) \
                .filter(Project.project_code == project_code) \
                .first()
            # And if we don't, throw an exception before we commit anything
            if project is not None:
                project.proprietary_duration = proprietary_duration
                if proprietary_duration != 0:
                    project.endtime = new_mjd_endtime
            else:
                raise UpdateException(f'Project {project_code} was not found in the archive')

            # We're not so lucky in the legacy archive, so we need to get all instances
            leg_project = legacy_session.query(LegacyProject) \
                .filter(LegacyProject.project_code == project_code) \
                .all()
            # Loop over each to set the period, or throw an exception if not found
            if len(leg_project) > 0:
                for project_instance in leg_project:
                    project_instance.proprietary_duration = proprietary_duration
                    project_instance.unlock_expire = proprietary_duration
                    if proprietary_duration != 0:
                        project_instance.proprietary = new_mjd_endtime
                        project_instance.project_lock = 'LTIME'
                    else:
                        project_instance.project_lock = 'PUBLIC'
            else:
                raise UpdateException(f'Project {project_code} was not found in the legacy archive')

            a_session.commit()
            a_session.close()
            legacy_session.commit()
            legacy_session.close()

    except Exception as update_exception:
        raise UpdateException(f'DB update failed for the following reason: {update_exception}')


def main(**kwargs):
    r"""
    The main entry point for this script.  Builds the parser, checks params, gets a profile and then
    attempts to update the db.  If that succeeds, we kick off an amygdala even to tell the system
    to re-index the project so the archive will reflect the new status.
    :param kwargs: command line arguments to be passed to our parser builder
    :return: nothing, if we complete the system will exit normally, if not we'll set a system
    exit code.
    """
    parser = _make_parser()
    args = parser.parse_args(**kwargs)

    capo_config = get_my_capo_config(profile=args.profile)

    if args.duration not in range(0, 730):
        _LOG.error(_DISALLOWED_DURATION)
        parser.print_help()
        sys.exit(2)

    try:
        set_project_proprietary_state(capo_config, args.project, args.duration)
    except UpdateException as update_exception:
        _LOG.error(update_exception)
        parser.print_help()
        sys.exit(3)

    # Set up a LogHandler to record the fact we just made a change to this project.
    # We're adding it here, instead of earlier, because nothing we log earlier should be presented
    # to anyone but the command line user and would only add useless clutter to our system logging.
    # We only really want the completed task to make a record in our system.
    broadcast = LogHandler(profile=capo_config.profile, application=_APPLICATION_NAME)
    broadcast.setLevel(logging.DEBUG)
    broadcast.setFormatter(LOG_MESSAGE_FORMATTER)
    _LOG.addHandler(broadcast)
    _LOG.info(f'Attempting to update proprietary period for {args.project}.')
    if args.duration != 0:
        _LOG.info(f'Locking for {args.duration} days from today')
    else:
        _LOG.info('Unlocking')

    event = {'logData': {'project_code': args.project,
                         'proprietary_duration': args.duration,
                         'ingestion_type': 'evla_sdm'
                         },
             'message': 'proprietary period updated',
             'request': 're-index please'}
    SendNRAOEvent(profile=capo_config.profile, application=_APPLICATION_NAME) \
        .send(routing_key='ingestion-complete.metadata', event=event)
    sys.exit(0)


if __name__ == '__main__':
    main()
