import logging
import sys
import warnings

from pyat import get_my_capo_config, get_console_logger
from pyat.schema import ArchiveDBSession, Project, Author, ScienceProduct
from sqlalchemy import exc as sa_exc, asc, desc

_APPLICATION_NAME = 'project_fetcher'
_LOG = get_console_logger(_APPLICATION_NAME, logging.DEBUG)
_MISSING_PROFILE = """ERROR: unknown 'profile'; provide the -P argument 
or set the CAPO_PROFILE environment variable."""
_DESCRIPTION = """A tool for retrieving project metadata from the archive."""

class ArchiveProjectFetcher:

    def __init__(self, profile):
        self.capo_config = get_my_capo_config(profile=profile)
        try:
            self.archive_context = ArchiveDBSession('SDM', profile=self.capo_config.profile)
            self.pst_context = ArchiveDBSession('PST', profile=self.capo_config.profile)
        except KeyError as k_ex:
            _LOG.error(f'An error occurred while creating a db context: {k_ex}')
            sys.exit(1)

    def fetch_project(self, project_code):
        with warnings.catch_warnings(), self.archive_context, self.pst_context:
            # Suppress SQLAlchemy warnings
            warnings.simplefilter("ignore", category=sa_exc.SAWarning)
            """
            Return the project specified by the input arguments, if it exists.
            :return: the first Project we found with the project code passed in
            """
            self.project = self.archive_context.session.query(Project) \
                .filter(Project.project_code == project_code) \
                .first()

            self.abstract = self.project.abstract

            self.is_alma = self._is_alma()
            self.authors = self._get_investigators()
            self.detachable_author_list = self._get_detached_authors()

            self.project_info = self.build_project_info()
            return self.project

    def build_project_info(self):
        output = []
        output.append(f'Title: {self.project.title}')
        output.append(f'Abstract: {self.project.abstract}')
        if self._is_alma():
            return

        investigator_list = self.authors
        # investigator_list = self._get_investigators()
        # We want the PI's pst_person_id followed by the CoIs' pst_person_ids in numeric order.
        # ALMA authors, however, do not have pst_person_ids
        pi = investigator_list[0]
        coi_pst_ids = [int(coi.pst_person_id) for coi in investigator_list[1:]]
        # TODO: should not need to sort; query does that
        coi_pst_ids = sorted(coi_pst_ids)

        author_pst_ids = [int(pi.pst_person_id)]
        [author_pst_ids.append(id) for id in coi_pst_ids]
        authors_to_print = [str(id) for id in author_pst_ids]
        id_list = ' '.join(authors_to_print)
        output.append(f'Authors: {id_list}')

        return output

    def _get_investigators(self):
        """
         Get a list of investigators associated with this project, with PI(s) as first element(s) of the list

         :return: a list of investigators associated with the project code passed in, ordered with
         the PI(s) first
        """
        investigators_list = self.archive_context.session.query(Author) \
            .filter(Author.project == self.project) \
            .order_by(desc(Author.is_pi), asc(Author.pst_person_id)) \
            .all()
        return investigators_list

    def _is_alma(self):
        with warnings.catch_warnings(), self.archive_context, self.pst_context:
            # Suppress SQLAlchemy warnings
            warnings.simplefilter("ignore", category=sa_exc.SAWarning)
            self.product = self.archive_context.session.query(ScienceProduct)\
                .filter(ScienceProduct.project == self.project)\
                .first()
            external_system = self.product.external_system
            return str(external_system).startswith("ALMA")

    def _get_detached_authors(self):
        return [DetachedAuthor(author) for author in self.authors]

if __name__ == '__main__':
    fetched = ArchiveProjectFetcher("nmtest").fetch_project('SK0442')
    authors = fetched.authors
    assert 4 == len(fetched.authors)
    _LOG.debug("looks ok")


class DetachedAuthor:

    def __init__(self, author:Author):
        self.author_id     = author.author_id
        self.username      = author.username
        self.firstname     = author.firstname
        self.lastname      = author.lastname
        self.pst_person_id = author.pst_person_id
        self.is_pi         = author.is_pi
        # self.project_code  = author.project_code

    def __eq__(self, other):
        if type(other) is not type(self):
            return False

        if other.author_id == self.author_id:
            if other.pst_person_id == self.pst_person_id:
                if other.lastname == self.lastname:
                    return other.firstname == self.firstname

        return False

    def __repr__(self):
        return f'{self.firstname} {self.lastname}: {self.author_id}, {self.pst_person_id}'